import datetime as dt
from smtplib import SMTPDataError, SMTPSenderRefused, SMTPRecipientsRefused

from flask import current_app
from flask_mail import Message, Connection
import sqlalchemy as sa
from werkzeug.local import LocalProxy


db = LocalProxy(lambda: current_app.extensions.get("sqlalchemy").db)
config = LocalProxy(lambda: current_app.config)
mail = LocalProxy(lambda: current_app.extensions.get("mail"))
mail_alchemy = LocalProxy(lambda: current_app.extensions.get("mail_alchemy"))


class MailNotSetError(Exception):
    pass


class MailLimitError(Exception):
    pass


class AttachmentMixin:
    """Declarative mixin class for e-mail attachments."""
    id = sa.Column(sa.Integer, autoincrement=True, primary_key=True)
    filename = sa.Column(sa.String(255))
    content_type = sa.Column(sa.String(40))
    data = sa.Column(sa.LargeBinary)


class EmailMixin:
    """Declarative mixin class for e-mail table."""
    id = sa.Column(sa.Integer, autoincrement=True, primary_key=True)
    sender_address = sa.Column(sa.String(255), nullable=False, index=True)
    sender_name = sa.Column(sa.String(80))
    subject = sa.Column(sa.String(80), nullable=False, index=True)
    recipient = sa.Column(sa.String(255), nullable=False, index=True)
    message_txt = sa.Column(sa.Text)
    message_html = sa.Column(sa.Text)

    scheduled_at = sa.Column(
        sa.DateTime,
        default=dt.datetime.utcnow,
        index=True
    )
    sent_at = sa.Column(
        sa.DateTime,
        index=True
    )
    error = sa.Column(sa.Text)

    @property
    def message(self) -> Message:
        sender = self.sender_address \
            if self.sender_name is None \
            else (self.sender_name, self.sender_address)
        msg = Message(
            self.subject,
            [self.recipient],
            self.message_txt,
            self.message_html,
            sender
        )

        if hasattr(self, "attachments"):
            for attachment in self.attachments:
                msg.attach(
                    filename=attachment.filename,
                    content_type=attachment.content_type,
                    data=attachment.data
                )

        return msg

    @classmethod
    def from_message(cls, msg: Message) -> tuple:
        messages = []
        if isinstance(msg.recipients, str):
            recipients = [msg.recipients]
        else:
            recipients = msg.recipients

        if msg.sender.endswith(">"):
            separator = msg.sender.rfind("<")
            sender_address = msg.sender[separator+1:-1]
            sender_name = msg.sender[:separator].strip()
        else:
            sender_address = msg.sender
            sender_name = None

        subject = msg.subject
        message_html = msg.html
        message_txt = msg.body

        attachments = []
        if hasattr(cls, "attachments"):
            for a in msg.attachments:
                attachment = mail_alchemy.attachment_class()
                attachment.filename = a.filename
                attachment.content_type = a.content_type
                attachment.data = a.data
                db.session.add(attachment)
                attachments.append(attachment)

        for recipient in recipients:
            email = cls()
            email.recipient = recipient
            email.sender_address = sender_address
            email.sender_name = sender_name
            email.subject = subject
            email.message_html = message_html
            email.message_txt = message_txt

            if hasattr(email, "attachments"):
                email.attachments.extend(attachments)

            messages.append(email)

        return tuple(messages)

    @classmethod
    def check_limit(cls):
        per_minute = config.get("MAIL_PER_MINUTE", None)
        per_hour = config.get("MAIL_PER_HOUR", None)
        per_day = config.get("MAIL_PER_DAY", None)

        now = dt.datetime.utcnow()

        qry = db.session.query(sa.func.count(cls.id))\
            .filter(cls.sent_at != None)\

        if per_minute is not None:
            in_minute = qry\
                .filter(cls.sent_at > now - dt.timedelta(minutes=1))\
                .first()[0] or 0
            if in_minute >= per_minute:
                raise MailLimitError(
                    f"Per minute limit ({per_minute}) exceeded."
                )

        if per_hour is not None:
            in_hour = qry\
                .filter(cls.sent_at > now - dt.timedelta(hours=1))\
                .first()\
                .first()[0] or 0
            if in_hour >= per_hour:
                raise MailLimitError(f"Hourly limit ({per_hour}) exceeded.")

        if per_day is not None:
            in_day = qry\
                .filter(cls.sent_at > now - dt.timedelta(days=1))\
                .first()\
                .first()[0] or 0
            if in_day >= per_day:
                raise MailLimitError(f"Daily limit ({per_day}) exceeded.")

    @classmethod
    def unsent(cls):
        qry = getattr(cls, 'query')
        return qry.filter_by(sent_at=None)

    @classmethod
    def scheduled(cls):
        return cls.unsent().filter(cls.scheduled_at < dt.datetime.utcnow())

    def send(self, connection: Connection = None):
        type(self).check_limit()

        msg = self.message

        try:
            if connection is not None:
                connection.send(msg)
            else:
                mail.send(msg)
            self.sent_at = dt.datetime.utcnow()
            db.session.commit()
        except (SMTPDataError, SMTPSenderRefused, SMTPRecipientsRefused) as e:
            self.sent_at = dt.datetime.utcnow()
            self.error = str(e)
            db.session.commit()
        except Exception as e:
            db.session.rollback()
            raise e

    @classmethod
    def send_scheduled(cls):
        with mail.connect() as conn:
            for email in cls.scheduled().all():
                try:
                    email.send(conn)
                except MailLimitError:
                    break
                except Exception:
                    continue



import datetime as dt
import threading as th
import time as tm

from flask import Flask, Blueprint, render_template
from flask_mail import Mail, Message
from flask_sqlalchemy import SQLAlchemy

from flask_mailalchemy.model import AttachmentMixin, EmailMixin


class MailAlchemy:
    """Flask-MailAlchemy extension

    Manages mail sending with scheduling and email sending limits.

    Args:
        app: Flask instance
        db: SQLAlchemy instance
        email_class: Email model class

    """
    def __init__(
            self,
            app: Flask = None,
            db: SQLAlchemy = None,
            email_class: type(EmailMixin) = None
    ):
        self.app = app
        self.db = db
        self.mail = None
        self.attachment_class = None
        self.email_class = email_class
        self.__stop_worker__ = False

        if app is not None:
            self.init_app(app, db)

    def init_app(
            self,
            app: Flask,
            db: SQLAlchemy,
            email_class: type(EmailMixin) = None
    ):
        """Initialize Flask-MailAlchemy

        Initializes Flask-Mail instance, creates email model class and registers
        blueprint `mail`.

        Args:
            app: Flask instance
            db: SQLAlchemy instance
            email_class: Email model class

        """
        self.app = app

        if db is None:
            db = app.extensions.get("sqlalchemy").db
        self.db = db

        self.mail = Mail(self.app)

        class Attachment(db.Model, AttachmentMixin):
            pass

        self.attachment_class = Attachment

        self.email_class = self.email_class or email_class
        if self.email_class is None:
            class Email(db.Model, EmailMixin):
                attachments = db.relationship(
                    Attachment,
                    secondary=db.Table(
                        "email_attachment",
                        db.Column(
                            "email_id",
                            db.Integer,
                            db.ForeignKey("email.id"),
                            primary_key=True
                        ),
                        db.Column(
                            "attachment_id",
                            db.Integer,
                            db.ForeignKey("attachment.id"),
                            primary_key=True
                        )
                    ),
                    backref="emails"
                )

            self.email_class = Email

        self.app.register_blueprint(
            Blueprint(
                "mail",
                __name__,
                url_prefix='/mail',
                template_folder="templates"
            )
        )

        self.app.extensions["mail_alchemy"] = self

    def send(self, msg: Message):
        """Sends a single message instance.

        Stores messages in database and sends them. A separate message is stored
        for every recipient in database. If TESTING is True the message will not
        actually be sent.

        Args:
            msg: a Message instance.
        """
        emails = self.email_class.from_message(msg)
        for email in emails:
            self.db.session.add(email)
            self.db.session.commit()

        for email in emails:
            email.send()
            self.db.session.commit()

    def send_message(self, *args, **kwargs):
        """Shortcut for send(msg).

        Takes same arguments as Message constructor.

        """
        msg = Message(*args, **kwargs)
        self.send(msg)

    @staticmethod
    def render_template(msg: Message, template: str, **context):
        """Renders plaintext and HTML content for Message.

        Message body is set from template found with .txt ending, html is set
        from template found with .html ending.

        Args:
            msg: Message instance
            template: Template name without extension
            **context: Template context

        """
        try:
            msg.body = render_template(
                "mail/{}.txt".format(template),
                **context
            )
        except FileNotFoundError:
            pass

        try:
            msg.html = render_template(
                "mail/{}.html".format(template),
                **context
            )
        except FileNotFoundError:
            pass

    def schedule(self, msg: Message, scheduled_at: dt.datetime = None):
        """Schedules a single message instance to send in future.

        Stores messages in database and sends them. A separate message is stored
        for every recipient in database.

        Args:
            msg: Message instance
            scheduled_at: Time of sending in future

        """
        if scheduled_at is None:
            scheduled_at = dt.datetime.utcnow()
        for email in self.email_class.from_message(msg):
            email.scheduled_at = scheduled_at
            self.db.session.add(email)

        self.db.session.commit()

    def schedule_message(
            self,
            *args,
            scheduled_at: dt.datetime = None,
            **kwargs
    ):
        """Shortcut for schedule(msg).

        Takes same arguments as Message constructor plus the `scheduled_at`
        datetime parameter.

        """
        msg = Message(*args, **kwargs)
        self.schedule(msg, scheduled_at=scheduled_at)

    def worker(self):
        """Sends unsent emails that are scheduled before current time."""
        with self.app.app_context():
            while not self.__stop_worker__:
                self.email_class.send_scheduled()
                tm.sleep(self.app.config.get("MAIL_ALCHEMY_CYCLE", 10))

    def run_worker(self):
        """Starts worker thread."""
        thread = th.Thread(target=self.worker)
        thread.start()

    def stop_worker(self):
        """Stops worker thread's loop."""
        self.__stop_worker__ = True



from setuptools import setup

# read the contents of your README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='Flask-MailAlchemy',
    version='0.4',
    packages=['flask_mailalchemy'],
    url='https://github.com/gerelorant/flask-mailalchemy',
    license='MIT',
    author='Gere Lóránt',
    author_email='gerelorant@gmail.com',
    description='SQLAlchemy based e-mail sending extension for Flask',
    install_requires=['Flask', "Flask-Mail", "Flask-SQLAlchemy"],
    include_package_data=True,
    long_description=long_description,
    long_description_content_type='text/markdown'
)
